import enum
import multiprocessing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Dict, Optional
import logging

import torch

from causy.serialization import SerializeMixin
from causy.graph_utils import load_pipeline_artefact_by_definition

logger = logging.getLogger(__name__)

DEFAULT_THRESHOLD = 0.01

AS_MANY_AS_FIELDS = 0


@dataclass
class ComparisonSettings(SerializeMixin):
    min: int = 2
    max: int = AS_MANY_AS_FIELDS


class NodeInterface(SerializeMixin):
    """
    Node interface for the graph. A node is defined by a name and a value.
    """

    name: str
    id: str
    values: torch.Tensor

    def serialize(self):
        return {"id": self.id, "name": self.name}


class EdgeTypeInterface:
    pass


class EdgeInterface(SerializeMixin):
    """
    Edge interface for the graph
    A graph edge is defined by two nodes and an edge type. It can also have metadata.
    """

    u: NodeInterface
    v: NodeInterface
    edge_type: EdgeTypeInterface
    metadata: Dict[str, any] = None

    def serialize(self):
        return {
            "edge_type": self.edge_type,
            "metadata": self.metadata,
        }


class TestResultAction(enum.StrEnum):
    """
    Actions that can be taken on the graph. These actions are used to keep track of the history of the graph.
    """

    REMOVE_EDGE_UNDIRECTED = "REMOVE_EDGE_UNDIRECTED"
    UPDATE_EDGE = "UPDATE_EDGE"
    UPDATE_EDGE_TYPE = "UPDATE_EDGE_TYPE"
    UPDATE_EDGE_DIRECTED = "UPDATE_EDGE_DIRECTED"
    UPDATE_EDGE_TYPE_DIRECTED = "UPDATE_EDGE_TYPE_DIRECTED"
    DO_NOTHING = "DO_NOTHING"
    REMOVE_EDGE_DIRECTED = "REMOVE_EDGE_DIRECTED"


@dataclass
class TestResult(SerializeMixin):
    u: NodeInterface
    v: NodeInterface
    action: TestResultAction
    data: Optional[Dict] = None

    def serialize(self):
        return {
            "from": self.u.serialize(),
            "to": self.v.serialize(),
            "action": self.action.name,
        }


class BaseGraphInterface(ABC):
    nodes: Dict[str, NodeInterface]
    edges: Dict[str, Dict[str, Dict]]

    @abstractmethod
    def retrieve_edge_history(self, u, v, action: TestResultAction) -> List[TestResult]:
        pass

    @abstractmethod
    def add_edge_history(self, u, v, action: TestResult):
        pass

    @abstractmethod
    def add_edge(self, u, v, metadata):
        pass

    @abstractmethod
    def remove_edge(self, u, v):
        pass

    @abstractmethod
    def remove_directed_edge(self, u, v):
        pass

    @abstractmethod
    def update_edge(self, u, v, metadata=None, edge_type=None):
        pass

    @abstractmethod
    def update_directed_edge(self, u, v, metadata=None, edge_type=None):
        pass

    @abstractmethod
    def add_node(self, name, values) -> NodeInterface:
        pass

    @abstractmethod
    def edge_value(self, u, v):
        pass

    @abstractmethod
    def undirected_edge_exists(self, u, v):
        pass

    @abstractmethod
    def directed_edge_exists(self, u, v):
        pass

    @abstractmethod
    def edge_exists(self, u, v):
        pass


class GraphModelInterface(ABC):
    pool: multiprocessing.Pool

    @abstractmethod
    def create_graph_from_data(self, data: List[Dict]):
        pass

    @abstractmethod
    def execute_pipeline_steps(self):
        pass

    @abstractmethod
    def execute_pipeline_step(self, step):
        pass


class GeneratorInterface(ABC, SerializeMixin):
    comparison_settings: ComparisonSettings
    chunked: bool = False

    @abstractmethod
    def generate(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        pass

    def __init__(self, comparison_settings: ComparisonSettings, chunked: bool = None):
        if isinstance(comparison_settings, dict):
            comparison_settings = load_pipeline_artefact_by_definition(
                comparison_settings
            )

        if chunked is not None:
            self.chunked = chunked

        self.comparison_settings = comparison_settings


class PipelineStepInterface(ABC, SerializeMixin):
    num_of_comparison_elements: int = 0
    generator: Optional[GeneratorInterface] = None

    chunk_size_parallel_processing: int = 1

    parallel: bool = True

    def __init__(
        self,
        threshold: float = DEFAULT_THRESHOLD,
        generator: Optional[GeneratorInterface] = None,
        num_of_comparison_elements: int = None,
        chunk_size_parallel_processing: int = None,
        parallel: bool = None,
    ):
        if generator:
            if isinstance(generator, dict):
                self.generator = load_pipeline_artefact_by_definition(generator)
            else:
                self.generator = generator

        if num_of_comparison_elements:
            if isinstance(num_of_comparison_elements, dict):
                self.num_of_comparison_elements = load_pipeline_artefact_by_definition(
                    num_of_comparison_elements
                )
            else:
                self.num_of_comparison_elements = num_of_comparison_elements

        if chunk_size_parallel_processing:
            self.chunk_size_parallel_processing = chunk_size_parallel_processing

        if parallel:
            self.parallel = parallel

        self.threshold = threshold

    @abstractmethod
    def test(self, nodes: List[str], graph: BaseGraphInterface) -> Optional[TestResult]:
        """
        Test if u and v are independent
        :param u: u values
        :param v: v values
        :return: True if independent, False otherwise
        """
        pass

    def __call__(
        self, nodes: List[str], graph: BaseGraphInterface
    ) -> Optional[TestResult]:
        return self.test(nodes, graph)


class LogicStepInterface(ABC, SerializeMixin):
    @abstractmethod
    def execute(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        pass


class ExitConditionInterface(ABC, SerializeMixin):
    @abstractmethod
    def check(
        self,
        graph: BaseGraphInterface,
        graph_model_instance_: GraphModelInterface,
        actions_taken: List[TestResult],
        iteration: int,
    ) -> bool:
        """
        :param graph:
        :param graph_model_instance_:
        :param actions_taken:
        :param iteration:
        :return: True if you want to break an iteration, False otherwise
        """
        pass

    def __call__(
        self,
        graph: BaseGraphInterface,
        graph_model_instance_: GraphModelInterface,
        actions_taken: List[TestResult],
        iteration: int,
    ) -> bool:
        return self.check(graph, graph_model_instance_, actions_taken, iteration)



from typing import Optional, List

from causy.interfaces import (
    LogicStepInterface,
    BaseGraphInterface,
    GraphModelInterface,
    PipelineStepInterface,
    ExitConditionInterface,
)
from causy.graph_utils import (
    load_pipeline_artefact_by_definition,
    load_pipeline_steps_by_definition,
)


class Loop(LogicStepInterface):
    """
    A loop which executes a list of pipeline_steps until the exit_condition is met.
    """

    def execute(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        """
        Executes the loop til self.exit_condition is met
        :param graph:
        :param graph_model_instance_:
        :return:
        """
        n = 0
        steps = None
        while not self.exit_condition(
            graph=graph,
            graph_model_instance_=graph_model_instance_,
            actions_taken=steps,
            iteration=n,
        ):
            steps = []
            for pipeline_step in self.pipeline_steps:
                result = graph_model_instance_.execute_pipeline_step(pipeline_step)
                steps.extend(result)
            n += 1

    def __init__(
        self,
        pipeline_steps: Optional[List[PipelineStepInterface]] = None,
        exit_condition: ExitConditionInterface = None,
    ):
        super().__init__()
        # TODO check if this is a good idea
        if isinstance(exit_condition, dict):
            exit_condition = load_pipeline_artefact_by_definition(exit_condition)

        # TODO: check if this is a good idea
        if len(pipeline_steps) > 0 and isinstance(pipeline_steps[0], dict):
            pipeline_steps = load_pipeline_steps_by_definition(pipeline_steps)

        self.pipeline_steps = pipeline_steps or []
        self.exit_condition = exit_condition



from causy.common_pipeline_steps.exit_conditions import ExitOnNoActions
from causy.generators import PairsWithNeighboursGenerator, RandomSampleGenerator
from causy.graph_model import graph_model_factory
from causy.common_pipeline_steps.logic import Loop
from causy.independence_tests.common import (
    CorrelationCoefficientTest,
    PartialCorrelationTest,
    ExtendedPartialCorrelationTestMatrix,
)
from causy.common_pipeline_steps.calculation import (
    CalculatePearsonCorrelations,
)
from causy.interfaces import AS_MANY_AS_FIELDS, ComparisonSettings
from causy.orientation_rules.pc import (
    ColliderTest,
    NonColliderTest,
    FurtherOrientTripleTest,
    OrientQuadrupleTest,
    FurtherOrientQuadrupleTest,
)

PC = graph_model_factory(
    pipeline_steps=[
        CalculatePearsonCorrelations(),
        CorrelationCoefficientTest(threshold=0.005),
        PartialCorrelationTest(threshold=0.005),
        ExtendedPartialCorrelationTestMatrix(threshold=0.005),
        ColliderTest(),
        Loop(
            pipeline_steps=[
                NonColliderTest(),
                FurtherOrientTripleTest(),
                OrientQuadrupleTest(),
                FurtherOrientQuadrupleTest(),
            ],
            exit_condition=ExitOnNoActions(),
        ),
    ]
)


ParallelPC = graph_model_factory(
    pipeline_steps=[
        CalculatePearsonCorrelations(),
        CorrelationCoefficientTest(threshold=0.005),
        PartialCorrelationTest(
            threshold=0.005, parallel=True, chunk_size_parallel_processing=50000
        ),
        ExtendedPartialCorrelationTestMatrix(
            # run first a sampled version of the test so we can minimize the number of tests in the full version
            threshold=0.005,
            chunk_size_parallel_processing=5000,
            parallel=True,
            generator=RandomSampleGenerator(
                generator=PairsWithNeighboursGenerator(
                    chunked=False,
                    shuffle_combinations=True,
                    comparison_settings=ComparisonSettings(
                        min=4, max=AS_MANY_AS_FIELDS
                    ),
                ),
                chunked=False,
                every_nth=200,
            ),
        ),
        ExtendedPartialCorrelationTestMatrix(
            threshold=0.005,
            chunk_size_parallel_processing=20000,
            parallel=True,
            generator=PairsWithNeighboursGenerator(
                chunked=False,
                shuffle_combinations=True,
                comparison_settings=ComparisonSettings(min=4, max=AS_MANY_AS_FIELDS),
            ),
        ),
        ColliderTest(),
        Loop(
            pipeline_steps=[
                NonColliderTest(),
                FurtherOrientTripleTest(),
                OrientQuadrupleTest(),
                FurtherOrientQuadrupleTest(),
            ],
            exit_condition=ExitOnNoActions(),
        ),
    ]
)



from typing import Tuple, List, Optional
import itertools

from causy.generators import AllCombinationsGenerator
from causy.interfaces import (
    BaseGraphInterface,
    TestResult,
    TestResultAction,
    PipelineStepInterface,
    ComparisonSettings,
)

# theory for all orientation rules with pictures:
# https://hpi.de/fileadmin/user_upload/fachgebiete/plattner/teaching/CausalInference/2019/Introduction_to_Constraint-Based_Causal_Structure_Learning.pdf

# TODO: refactor ColliderTest -> ColliderRule and move to folder orientation_rules (after checking for duplicates)


class ColliderTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        We call triples u, v, z of nodes v structures if u and v that are NOT adjacent but share an adjacent node z.
        V structures looks like this in the undirected skeleton: (u - z - v).
        We now check if z is in the separating set.
        If z is not in the separating set, we know that u and v are uncorrelated given z.
        So, the edges must be oriented from u to z and from v to z (u -> z <- v).
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """
        # https://github.com/pgmpy/pgmpy/blob/1fe10598df5430295a8fc5cdca85cf2d9e1c4330/pgmpy/estimators/PC.py#L416

        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        # if u and v are adjacent, do nothing
        if graph.undirected_edge_exists(x, y):
            return TestResult(u=x, v=y, action=TestResultAction.DO_NOTHING, data={})

        # if u and v are NOT adjacent, store all shared adjacent nodes
        potential_zs = set(graph.edges[x.id].keys()).intersection(
            set(graph.edges[y.id].keys())
        )

        actions = graph.retrieve_edge_history(
            x, y, TestResultAction.REMOVE_EDGE_UNDIRECTED
        )

        # if u and v are not independent given z, safe action: make z a collider
        results = []
        for z in potential_zs:
            z = graph.nodes[z]

            separators = []
            for action in actions:
                if "separatedBy" in action.data:
                    separators += [a.id for a in action.data["separatedBy"]]

            if z.id not in separators:
                results += [
                    TestResult(
                        u=z,
                        v=x,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    ),
                    TestResult(
                        u=z,
                        v=y,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    ),
                ]
        return results


class NonColliderTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        Further orientation rule: all v structures that are colliders are already oriented.
        We now orient all v structures that have a single alternative to being a collider.
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """

        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        # if u and v are adjacent, do nothing
        if graph.edge_exists(x, y):
            return

        # if u and v are NOT adjacent, store all shared adjacent nodes
        potential_zs = set(graph.edges[x.id].keys()).intersection(
            set(graph.edges[y.id].keys())
        )
        # if one edge has an arrowhead at z, orient the other one pointing away from z.
        # It cannot be a collider because we have already oriented all unshielded triples that contain colliders.
        for z in potential_zs:
            z = graph.nodes[z]
            breakflag = False
            if graph.only_directed_edge_exists(x, z) and graph.undirected_edge_exists(
                z, y
            ):
                for node in graph.nodes:
                    if graph.only_directed_edge_exists(graph.nodes[node], y):
                        breakflag = True
                        break
                if breakflag is True:
                    continue
                return TestResult(
                    u=y,
                    v=z,
                    action=TestResultAction.REMOVE_EDGE_DIRECTED,
                    data={},
                )

            if graph.only_directed_edge_exists(y, z) and graph.undirected_edge_exists(
                z, x
            ):
                for node in graph.nodes:
                    if graph.only_directed_edge_exists(graph.nodes[node], x):
                        continue
                return TestResult(
                    u=x,
                    v=z,
                    action=TestResultAction.REMOVE_EDGE_DIRECTED,
                    data={},
                )


class FurtherOrientTripleTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        Further orientation rule.
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """

        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        results = []
        for z in graph.nodes:
            z = graph.nodes[z]
            # check if it is a potential z
            if not (graph.edge_exists(y, z) and graph.edge_exists(x, z)):
                continue

            if (
                graph.undirected_edge_exists(x, y)
                and graph.only_directed_edge_exists(x, z)
                and graph.only_directed_edge_exists(z, y)
            ):
                results.append(
                    TestResult(
                        u=y,
                        v=x,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
            if (
                graph.undirected_edge_exists(x, y)
                and graph.only_directed_edge_exists(y, z)
                and graph.only_directed_edge_exists(z, x)
            ):
                results.append(
                    TestResult(
                        u=x,
                        v=y,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
        return results


class OrientQuadrupleTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        Further orientation rule.
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """

        y = graph.nodes[nodes[0]]
        w = graph.nodes[nodes[1]]

        potential_zs = set()

        # TODO: just iterate over edges
        for z in graph.nodes:
            z = graph.nodes[z]
            if graph.edge_exists(y, z) and graph.edge_exists(z, w):
                potential_zs.add(z)

        results = []
        for zs in itertools.combinations(potential_zs, 2):
            x, z = zs
            if (
                not graph.edge_exists(y, w)
                and graph.only_directed_edge_exists(y, z)
                and graph.only_directed_edge_exists(w, z)
                and graph.undirected_edge_exists(x, y)
                and graph.undirected_edge_exists(x, w)
                and graph.undirected_edge_exists(x, z)
            ):
                results.append(
                    TestResult(
                        u=z,
                        v=x,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
            if (
                not graph.edge_exists(y, w)
                and graph.only_directed_edge_exists(y, x)
                and graph.only_directed_edge_exists(w, x)
                and graph.undirected_edge_exists(y, z)
                and graph.undirected_edge_exists(w, z)
                and graph.undirected_edge_exists(x, z)
            ):
                results.append(
                    TestResult(
                        u=x,
                        v=z,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
        return results


class FurtherOrientQuadrupleTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        Further orientation rule.
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """

        x = graph.nodes[nodes[0]]
        w = graph.nodes[nodes[1]]

        potential_zs = set()

        # TODO: just iterate over edges
        for z in graph.nodes:
            z = graph.nodes[z]
            if graph.edge_exists(x, z) and graph.edge_exists(z, w):
                potential_zs.add(z)

        results = []
        for zs in itertools.combinations(potential_zs, 2):
            y, z = zs
            if (
                not graph.edge_exists(y, z)
                and graph.undirected_edge_exists(x, z)
                and graph.undirected_edge_exists(x, w)
                and graph.undirected_edge_exists(x, y)
                and graph.only_directed_edge_exists(y, w)
                and graph.only_directed_edge_exists(w, z)
            ):
                results.append(
                    TestResult(
                        u=z,
                        v=x,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
            elif (
                not graph.edge_exists(z, y)
                and graph.undirected_edge_exists(x, y)
                and graph.undirected_edge_exists(x, w)
                and graph.undirected_edge_exists(x, z)
                and graph.only_directed_edge_exists(z, w)
                and graph.only_directed_edge_exists(w, y)
            ):
                results.append(
                    TestResult(
                        u=y,
                        v=x,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
            elif (
                not graph.edge_exists(y, z)
                and graph.undirected_edge_exists(w, z)
                and graph.undirected_edge_exists(x, w)
                and graph.undirected_edge_exists(w, y)
                and graph.only_directed_edge_exists(y, x)
                and graph.only_directed_edge_exists(x, z)
            ):
                results.append(
                    TestResult(
                        u=z,
                        v=w,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
            elif (
                not graph.edge_exists(z, y)
                and graph.undirected_edge_exists(w, y)
                and graph.undirected_edge_exists(x, w)
                and graph.undirected_edge_exists(w, z)
                and graph.only_directed_edge_exists(z, x)
                and graph.only_directed_edge_exists(x, y)
            ):
                results.append(
                    TestResult(
                        u=y,
                        v=w,
                        action=TestResultAction.REMOVE_EDGE_DIRECTED,
                        data={},
                    )
                )
        return results



import logging
from abc import ABC
from copy import deepcopy
from typing import Optional, List, Dict

import torch.multiprocessing as mp

from causy.graph import Graph, EdgeType
from causy.graph_utils import unpack_run
from causy.interfaces import (
    PipelineStepInterface,
    TestResultAction,
    LogicStepInterface,
    BaseGraphInterface,
    GraphModelInterface,
)

logger = logging.getLogger(__name__)


class AbstractGraphModel(GraphModelInterface, ABC):
    """
    The graph model is the main class of causy. It is responsible for creating a graph from data and executing the pipeline_steps.

    The graph model is responsible for the following tasks:
    - Create a graph from data (create_graph_from_data)
    - Execute the pipeline_steps (execute_pipeline_steps)
    - Take actions on the graph (execute_pipeline_step & _take_action which is called by execute_pipeline_step)

    It also initializes and takes care of the multiprocessing pool.

    """

    pipeline_steps: List[PipelineStepInterface]
    graph: BaseGraphInterface
    pool: mp.Pool

    def __init__(
        self,
        graph=None,
        pipeline_steps: Optional[List[PipelineStepInterface]] = None,
    ):
        self.graph = graph
        self.pipeline_steps = pipeline_steps or []
        self.pool = mp.Pool(mp.cpu_count() * 2)

    def create_graph_from_data(self, data: List[Dict[str, float]]):
        """
        Create a graph from data
        :param data: is a list of dictionaries
        :return:
        """
        # initialize nodes
        keys = data[0].keys()
        nodes: Dict[str, List[float]] = {}

        for key in keys:
            nodes[key] = []

        # load nodes into node dict
        for row in data:
            for key in keys:
                nodes[key].append(row[key])

        graph = Graph()
        for key in keys:
            graph.add_node(key, nodes[key])

        self.graph = graph
        return graph

    def create_all_possible_edges(self):
        """
        Create all possible edges on a graph
        TODO: replace me with the skeleton builders
        :return:
        """
        for u in self.graph.nodes.values():
            for v in self.graph.nodes.values():
                if u == v:
                    continue
                self.graph.add_edge(u, v, {})

    def execute_pipeline_steps(self):
        """
        Execute all pipeline_steps
        :return: the steps taken during the step execution
        """
        action_history = []

        for filter in self.pipeline_steps:
            logger.info(f"Executing pipeline step {filter.__class__.__name__}")
            if isinstance(filter, LogicStepInterface):
                filter.execute(self.graph, self)
                continue

            result = self.execute_pipeline_step(filter)
            action_history.append(
                {"step": filter.__class__.__name__, "actions": result}
            )

        return action_history

    def _format_yield(self, test_fn, graph, generator):
        """
        Format the yield for the parallel processing
        :param test_fn: the pipeline_step test function
        :param graph: the graph
        :param generator: the generator object which generates the combinations
        :return: yields the test function with its inputs
        """
        for i in generator:
            yield [test_fn, [*i], graph]

    def _take_action(self, results):
        """
        Take the actions returned by the test

        In causy changes on the graph are not executed directly. Instead, the test returns an action which should be executed on the graph.
        This is done to make it possible to execute the tests in parallel as well as to decide proactively at which point in the decisions taken by the pipeline step should be executed.
        Actions are returned by the test and are executed on the graph. The actions are stored in the action history to make it possible to revert the actions or use them in a later step.

        :param results:
        :return:
        """
        actions_taken = []
        for result_items in results:
            if result_items is None:
                continue
            if not isinstance(result_items, list):
                result_items = [result_items]

            for i in result_items:
                if i.u is not None and i.v is not None:
                    logger.info(f"Action: {i.action} on {i.u.name} and {i.v.name}")

                # execute the action returned by the test
                if i.action == TestResultAction.REMOVE_EDGE_UNDIRECTED:
                    if not self.graph.undirected_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to remove undirected edge {i.u.name} <-> {i.v.name}. But it does not exist."
                        )
                        continue
                    self.graph.remove_edge(i.u, i.v)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)
                elif i.action == TestResultAction.UPDATE_EDGE:
                    if not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self.graph.update_edge(i.u, i.v, metadata=i.data)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)
                elif i.action == TestResultAction.UPDATE_EDGE_DIRECTED:
                    if not self.graph.directed_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update directed edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self.graph.update_directed_edge(i.u, i.v, i.data)
                    self.graph.add_edge_history(i.u, i.v, i)
                elif i.action == TestResultAction.DO_NOTHING:
                    continue
                elif i.action == TestResultAction.REMOVE_EDGE_DIRECTED:
                    if not self.graph.directed_edge_exists(
                        i.u, i.v
                    ) and not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to remove directed edge {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue

                    self.graph.remove_directed_edge(i.u, i.v)
                    # TODO: move this to pre/post update hooks
                    if self.graph.edge_exists(
                        i.v, i.u
                    ):  # if the edge is undirected, make it directed
                        self.graph.update_directed_edge(
                            i.v, i.u, edge_type=EdgeType.DIRECTED
                        )
                    self.graph.add_edge_history(i.u, i.v, i)

                elif i.action == TestResultAction.UPDATE_EDGE_TYPE:
                    if not self.graph.edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge type {i.u.name} <-> {i.v.name}. But it does not exist."
                        )
                        continue
                    self.graph.update_edge(i.u, i.v, edge_type=i.edge_type)
                    self.graph.add_edge_history(i.u, i.v, i)
                    self.graph.add_edge_history(i.v, i.u, i)

                elif i.action == TestResultAction.UPDATE_EDGE_TYPE_DIRECTED:
                    if not self.graph.directed_edge_exists(i.u, i.v):
                        logger.debug(
                            f"Tried to update edge type {i.u.name} -> {i.v.name}. But it does not exist."
                        )
                        continue
                    self.graph.update_directed_edge(i.u, i.v, edge_type=i.edge_type)
                    self.graph.add_edge_history(i.u, i.v, i)

                # add the action to the actions history
                actions_taken.append(i)

        return actions_taken

    def execute_pipeline_step(self, test_fn: PipelineStepInterface):
        """
        Execute a single pipeline_step on the graph. either in parallel or in a single process depending on the test_fn.parallel flag
        :param test_fn: the test function
        :param threshold: the threshold
        :return:
        """
        actions_taken = []

        # initialize the worker pool (we currently use all available cores * 2)

        # run all combinations in parallel except if the number of combinations is smaller then the chunk size
        # because then we would create more overhead then we would definetly gain from parallel processing
        if test_fn.parallel:
            for result in self.pool.imap_unordered(
                unpack_run,
                self._format_yield(
                    test_fn, self.graph, test_fn.generator.generate(self.graph, self)
                ),
                chunksize=test_fn.chunk_size_parallel_processing,
            ):
                if not isinstance(result, list):
                    result = [result]
                actions_taken.extend(self._take_action(result))
        else:
            if test_fn.generator.chunked:
                for chunk in test_fn.generator.generate(self.graph, self):
                    iterator = [
                        unpack_run(i)
                        for i in [[test_fn, [*c], self.graph] for c in chunk]
                    ]
                    actions_taken.extend(self._take_action(iterator))
            else:
                iterator = [
                    unpack_run(i)
                    for i in [
                        [test_fn, [*i], self.graph]
                        for i in test_fn.generator.generate(self.graph, self)
                    ]
                ]
                actions_taken.extend(self._take_action(iterator))

        self.graph.action_history.append(
            {"step": type(test_fn).__name__, "actions": actions_taken}
        )

        return actions_taken


def graph_model_factory(
    pipeline_steps: Optional[List[PipelineStepInterface]] = None,
) -> type[AbstractGraphModel]:
    """
    Create a graph model based on a List of pipeline_steps
    :param pipeline_steps: a list of pipeline_steps which should be applied to the graph
    :return: the graph model
    """

    class GraphModel(AbstractGraphModel):
        def __init__(self):
            super().__init__(pipeline_steps=pipeline_steps)

    return GraphModel



from causy.graph_utils import serialize_module_name


def serialize_model(model, algorithm_name: str = None):
    """Serialize the model into a dictionary."""
    output = []
    for step in model.pipeline_steps:
        output.append(step.serialize())

    return {"name": algorithm_name, "steps": output}


class SerializeMixin:
    """Mixin class for serializing and deserializing graph steps."""

    def _serialize_object(self, obj):
        """Serialize the object into a dictionary."""
        result = {}
        for attr in [
            attr
            for attr in dir(obj)
            if not attr.startswith("__") and not attr.startswith("_")
        ]:
            if type(getattr(self, attr)) in [int, float, str, bool, type(None)]:
                result[attr] = getattr(self, attr)
            elif isinstance(getattr(self, attr), SerializeMixin):
                result[attr] = getattr(self, attr).serialize()
            elif isinstance(getattr(self, attr), list):
                result[attr] = [
                    x.serialize() if isinstance(x, SerializeMixin) else x
                    for x in getattr(self, attr)
                ]
            elif isinstance(getattr(self, attr), dict):
                result[attr] = {
                    x.serialize() if isinstance(x, SerializeMixin) else x
                    for x in getattr(self, attr)
                }
            elif isinstance(getattr(self, attr), tuple) or isinstance(
                getattr(self, attr), set
            ):
                # tuples are immutable, so we have to convert them to lists
                result[attr] = [
                    x.serialize() if isinstance(x, SerializeMixin) else x
                    for x in getattr(self, attr)
                ]

        return result

    def serialize(self):
        """Serialize the object into a dictionary."""
        # get all attributes of the class and its children
        params = self._serialize_object(self.__class__)
        params.update(self._serialize_object(self))

        return {
            "name": serialize_module_name(self),
            "params": params,
        }



import importlib
import json
from datetime import datetime
from json import JSONEncoder
import logging

import typer

from causy.graph_model import graph_model_factory
from causy.serialization import serialize_model
from causy.graph_utils import (
    load_pipeline_steps_by_definition,
    retrieve_edges,
)
from causy.ui import server

app = typer.Typer()


def load_json(pipeline_file: str):
    with open(pipeline_file, "r") as file:
        pipeline = json.loads(file.read())
    return pipeline


def load_algorithm(algorithm: str):
    st_function = importlib.import_module("causy.algorithms")
    st_function = getattr(st_function, algorithm)
    if not st_function:
        raise ValueError(f"Algorithm {algorithm} not found")
    return st_function


def create_pipeline(pipeline_config: dict):
    return load_pipeline_steps_by_definition(pipeline_config["steps"])


class MyJSONEncoder(JSONEncoder):
    def default(self, obj):
        return obj.serialize()


@app.command()
def eject(algorithm: str, output_file: str):
    typer.echo(f"💾 Loading algorithm {algorithm}")
    model = load_algorithm(algorithm)()
    result = serialize_model(model, algorithm_name=algorithm)
    typer.echo(f"💾 Saving algorithm {algorithm} to {output_file}")
    with open(output_file, "w") as file:
        file.write(json.dumps(result, indent=4))


@app.command()
def ui(result_file: str):
    result = load_json(result_file)

    server_config, server_runner = server(result)
    typer.launch(f"http://{server_config.host}:{server_config.port}")
    typer.echo(f"🚀 Starting server at http://{server_config.host}:{server_config.port}")
    server_runner.run()


@app.command()
def execute(
    data_file: str,
    pipeline: str = None,
    algorithm: str = None,
    output_file: str = None,
    log_level: str = "ERROR",
):
    logging.basicConfig(level=log_level)
    if pipeline:
        typer.echo(f"💾 Loading pipeline from {pipeline}")
        pipeline_config = load_json(pipeline)
        pipeline = create_pipeline(pipeline_config)
        model = graph_model_factory(pipeline_steps=pipeline)()
        algorithm_reference = {
            "type": "pipeline",
            "reference": pipeline,  # TODO: how to reference pipeline in a way that it can be loaded?
        }
    elif algorithm:
        typer.echo(f"💾 Creating pipeline from algorithm {algorithm}")
        model = load_algorithm(algorithm)()
        algorithm_reference = {
            "type": "default",
            "reference": algorithm,
        }

    else:
        raise ValueError("Either pipeline_file or algorithm must be specified")

    # initialize from json
    model.create_graph_from_data(load_json(data_file))

    # TODO: I should become a configurable skeleton builder
    model.create_all_possible_edges()

    typer.echo("🕵🏻‍♀  Executing pipeline steps...")
    model.execute_pipeline_steps()
    edges = []
    for edge in retrieve_edges(model.graph):
        print(
            f"{model.graph.nodes[edge[0]].name} -> {model.graph.nodes[edge[1]].name}: {model.graph.edges[edge[0]][edge[1]]}"
        )
        edges.append(
            {
                "from": model.graph.nodes[edge[0]].serialize(),
                "to": model.graph.nodes[edge[1]].serialize(),
                "value": model.graph.edges[edge[0]][edge[1]],
            }
        )

    if output_file:
        typer.echo(f"💾 Saving graph actions to {output_file}")
        with open(output_file, "w") as file:
            export = {
                "name": algorithm,
                "created_at": datetime.now().isoformat(),
                "algorithm": algorithm_reference,
                "steps": model.graph.action_history,
                "nodes": model.graph.nodes,
                "edges": edges,
            }
            file.write(json.dumps(export, cls=MyJSONEncoder, indent=4))


if __name__ == "__main__":
    app()



import itertools
from typing import Tuple, List, Optional
import logging

import torch

from causy.generators import AllCombinationsGenerator, PairsWithNeighboursGenerator
from causy.math_utils import get_t_and_critical_t
from causy.interfaces import (
    PipelineStepInterface,
    BaseGraphInterface,
    NodeInterface,
    TestResult,
    TestResultAction,
    AS_MANY_AS_FIELDS,
    ComparisonSettings,
)

logger = logging.getLogger(__name__)


class CorrelationCoefficientTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(self, nodes: List[str], graph: BaseGraphInterface) -> Optional[TestResult]:
        """
        Test if u and v are independent and delete edge in graph if they are.
        :param nodes: list of nodes
        :return: A TestResult with the action to take
        """
        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        # make t test for independency of u and v
        sample_size = len(x.values)
        nb_of_control_vars = 0
        corr = graph.edge_value(x, y)["correlation"]
        t, critical_t = get_t_and_critical_t(
            sample_size, nb_of_control_vars, corr, self.threshold
        )
        if abs(t) < critical_t:
            logger.debug(f"Nodes {x.name} and {y.name} are uncorrelated")
            return TestResult(
                u=x,
                v=y,
                action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                data={},
            )


class PartialCorrelationTest(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=3, max=3)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult]]:
        """
        Test if nodes u,v are independent given node z based on a partial correlation test.
        We use this test for all combinations of 3 nodes because it is faster than the extended test (which supports combinations of n nodes). We can
        use it to remove edges between nodes which are not independent given another node and so reduce the number of combinations for the extended test.
        :param nodes: the nodes to test
        :return: A TestResult with the action to take

        TODO: we are testing (C and E given B) and (E and C given B), we just need one of these, remove redundant tests.
        """
        results = []
        already_deleted_edges = set()
        for nodes in itertools.permutations(nodes):
            x: NodeInterface = graph.nodes[nodes[0]]
            y: NodeInterface = graph.nodes[nodes[1]]
            z: NodeInterface = graph.nodes[nodes[2]]

            # Avoid division by zero
            if x is None or y is None or z is None:
                return

            if not graph.edge_exists(x, y) or (y, x) in already_deleted_edges:
                continue

            try:
                cor_xy = graph.edge_value(x, y)["correlation"]
                cor_xz = graph.edge_value(x, z)["correlation"]
                cor_yz = graph.edge_value(y, z)["correlation"]
            except (KeyError, TypeError):
                return

            numerator = cor_xy - cor_xz * cor_yz
            denominator = ((1 - cor_xz**2) * (1 - cor_yz**2)) ** 0.5

            # Avoid division by zero
            if denominator == 0:
                return

            par_corr = numerator / denominator

            # make t test for independency of u and v given z
            sample_size = len(x.values)
            nb_of_control_vars = len(nodes) - 2
            t, critical_t = get_t_and_critical_t(
                sample_size, nb_of_control_vars, par_corr, self.threshold
            )

            if abs(t) < critical_t:
                logger.debug(
                    f"Nodes {x.name} and {y.name} are uncorrelated given {z.name}"
                )

                results.append(
                    TestResult(
                        u=x,
                        v=y,
                        action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                        data={"separatedBy": [z]},
                    )
                )
                already_deleted_edges.add((x, y))
        return results


class ExtendedPartialCorrelationTestMatrix(PipelineStepInterface):
    generator = PairsWithNeighboursGenerator(
        comparison_settings=ComparisonSettings(min=4, max=AS_MANY_AS_FIELDS)
    )
    chunk_size_parallel_processing = 1000
    parallel = False

    def test(self, nodes: List[str], graph: BaseGraphInterface) -> Optional[TestResult]:
        """
        Test if nodes u,v are independent given Z (set of nodes) based on partial correlation using the inverted covariance matrix (precision matrix).
        https://en.wikipedia.org/wiki/Partial_correlation#Using_matrix_inversion
        We use this test for all combinations of more than 3 nodes because it is slower.
        :param nodes: the nodes to test
        :return: A TestResult with the action to take
        """

        if not graph.edge_exists(graph.nodes[nodes[0]], graph.nodes[nodes[1]]):
            return

        other_neighbours = set(graph.edges[nodes[0]])
        other_neighbours.remove(graph.nodes[nodes[1]].id)

        if not set(nodes[2:]).issubset(set([on for on in list(other_neighbours)])):
            return

        inverse_cov_matrix = torch.inverse(
            torch.cov(torch.stack([graph.nodes[node].values for node in nodes]))
        )
        n = inverse_cov_matrix.size(0)
        diagonal = torch.diag(inverse_cov_matrix)
        diagonal_matrix = torch.zeros((n, n), dtype=torch.float32)
        for i in range(n):
            diagonal_matrix[i, i] = diagonal[i]

        helper = torch.mm(torch.sqrt(diagonal_matrix), inverse_cov_matrix)
        precision_matrix = torch.mm(helper, torch.sqrt(diagonal_matrix))

        sample_size = len(graph.nodes[nodes[0]].values)
        nb_of_control_vars = len(nodes) - 2

        t, critical_t = get_t_and_critical_t(
            sample_size,
            nb_of_control_vars,
            (
                (-1 * precision_matrix[0][1])
                / torch.sqrt(precision_matrix[0][0] * precision_matrix[1][1])
            ).item(),
            self.threshold,
        )

        if abs(t) < critical_t:
            logger.debug(
                f"Nodes {graph.nodes[nodes[0]].name} and {graph.nodes[nodes[1]].name} are uncorrelated given nodes {','.join([graph.nodes[on].name for on in other_neighbours])}"
            )
            nodes_set = set([graph.nodes[n] for n in nodes])
            return TestResult(
                u=graph.nodes[nodes[0]],
                v=graph.nodes[nodes[1]],
                action=TestResultAction.REMOVE_EDGE_UNDIRECTED,
                data={
                    "separatedBy": list(
                        nodes_set - {graph.nodes[nodes[0]], graph.nodes[nodes[1]]}
                    )
                },
            )



import abc
from types import SimpleNamespace
import torch

from causy.graph import Graph
from typing import Dict, Callable, List, Optional, Union

import logging

logger = logging.getLogger(__name__)


def random() -> torch.Tensor:
    """
    Returns a random number from a normal distribution
    :return: the random number as a float
    """
    return torch.randn(1, dtype=torch.float32).item()


class SampleEdge:
    """
    Represents an edge in a sample graph
    defines a directed edge from source to target

    Example:
    SampleEdge("X", "Y") defines an edge from X to Y

    """

    def __init__(self, source, target):
        self.source = source
        self.target = target


class SampleLaggedEdge(SampleEdge):
    """
    Represents a lagged edge in a time series sample graph.
    Defines a directed edge from source to target with a lag of lag. The lag is the number of time steps between the
    source and the target.

    Example:
    SampleLaggedEdge("X", "Y", 1) defines an edge from X to Y with a lag of 1

    """

    def __init__(self, source, target, lag):
        super().__init__(source, target)
        self.lag = lag


class TimeTravelingError(Exception):
    """
    An error that is raised when a TimeProxy tries to access a value from the future
    """

    pass


class TimeProxy:
    """
    A proxy object that allows to access past values of a variable by using the t() method.

    Example:
    >>> tp = TimeProxy(5)
    >>> tp.set_current_time(1)
    >>> tp.t(-1)
    5
    """

    def __init__(self, initial_value: torch.Tensor):
        """
        :param initial_value: the initial value of the variable
        """
        if not isinstance(initial_value, torch.Tensor):
            initial_value = torch.tensor(initial_value, dtype=torch.float32)
        self._lst = [initial_value]
        self._t = 0

    def set_current_time(self, t):
        """
        Set the current time step which t will be relative to
        :param t: time step as an integer
        :return:
        """
        self._t = t

    def t(self, t):
        """
        Return the value of the variable at time step t
        :param t: the relative time step as a negative integer
        :return: the value of the variable at time step t or a random number if t is too large
        :raises TimeTravelingError: if t is positive
        """
        if t > 0:
            raise TimeTravelingError(
                f"Cannot travel into the future ({self._t + t} steps ahead)."
            )
        elif t + self._t < 0:
            # TODO this might be a bad idea
            return random()
        return self._lst[self._t + t]

    def append(self, value):
        """
        Append a value to the list of values
        :param value: the value to append
        :return:
        """
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value, dtype=torch.float32)
        self._lst.append(value)

    def to_list(self):
        """
        Return the list of values
        :return: the list
        """
        return torch.stack(self._lst)

    def __repr__(self):
        return f"{self._lst} at {self._t}"


class CurrentElementProxy(float):
    """
    A proxy object that allows to access the current value of a variable. It is a subclass of float, so it can be used
    as a float.
    """

    # TODO: fix this class. Bug: IIDSampleGenerator only depends on the initial value, it should depend on the step

    def __init__(self, initial_value: torch.Tensor):
        if not isinstance(initial_value, torch.Tensor):
            initial_value = torch.tensor(initial_value, dtype=torch.float32)
        self.lst = [initial_value]
        self.value = initial_value

    def append(self, value):
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value, dtype=torch.float32)

        self.lst.append(value)
        self.value = value

    def to_list(self):
        return torch.stack(self.lst)


class AbstractSampleGenerator(abc.ABC):
    """
    An abstract class for sample generators that generate data for a sample graph.
    It is implemented by TimeseriesSampleGenerator and IIDSampleGenerator.

    """

    def __init__(
        self,
        initial_values: Dict[str, any],
        generators: Dict[str, Callable],
        edges: List[Union[SampleLaggedEdge, SampleEdge]],
        variables: Optional[Dict[str, any]] = None,
    ):
        self.initial_values = initial_values
        self.generators = generators
        self.edges = edges
        if variables is None:
            variables = {}
        self.vars = variables

    @abc.abstractmethod
    def generate(self, size: int):
        pass


class TimeseriesSampleGenerator(AbstractSampleGenerator):
    """
    A sample generator that generates data for a sample graph with a time dimension.

    Generator functions are written as a lambda function that takes two arguments: the current timestep and the input.
    The input is a SimpleNamespace object that contains the current values of all variables. Via the t(-n) method of the
    TimeProxy class, the generator can access the values of the variables at previous time steps.

    During generation, the generator functions are called in the order of the keys of the generators dictionary. This
    means that the order of the keys determines the order in which the variables are generated. If a variable depends on
    another variable, the generator function of the dependent variable should be defined after the generator function of
    the variable it depends on.

    A variable can depend on itself, but only on its past values (with a lag). This is useful for autoregressive models.

    Example:
    >>> sg = TimeseriesSampleGenerator(
    >>>     initial_values={
    >>>         "Z": random(),
    >>>         "Y": random(),
    >>>         "X": random(),
    >>>     },
    >>>     variables={
    >>>         "alpha": 0.9,
    >>>     },
    >>>     # generate the dependencies of variables on past values of themselves and other variables
    >>>     generators={
    >>>         "Z": lambda t, i: i.Z.t(-1) * i.alpha + random(),
    >>>         "Y": lambda t, i: i.Y.t(-1) * i.alpha + i.Z.t(-1) + random(),
    >>>         "X": lambda t, i: i.X.t(-1) * i.alpha + i.Y.t(-1) + random()
    >>>     },
    >>>     edges=[
    >>>         SampleLaggedEdge("X", "X", 1),
    >>>         SampleLaggedEdge("Y", "Y", 1),
    >>>         SampleLaggedEdge("Z", "Z", 1),
    >>>         SampleLaggedEdge("Y", "Z", 1),
    >>>         SampleLaggedEdge("Y", "X", 4),
    >>>     ]
    >>> )
    """

    def _generate_data(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size:
        :return:
        """
        internal_repr = {}

        # Initialize the output dictionary
        for i, v in self.initial_values.items():
            internal_repr[i] = TimeProxy(v)

        # Generate the data for each time step
        for t in range(1, size):
            for value in self.generators.keys():
                internal_repr[value].set_current_time(t)
                generator_input = internal_repr
                generator_input.update(self.vars)
                internal_repr[value].append(
                    self.generators[value](t, SimpleNamespace(**generator_input))
                )

        output = {}
        for i in self.generators.keys():
            output[i] = internal_repr[i].to_list()

        return output

    def generate(self, size):
        """
        Generate data for a sample graph with a time dimension
        :param size: the number of time steps to generate
        :return: the generated data and the sample graph
        """
        output = self._generate_data(size)
        graph = Graph()
        for i in self.generators.keys():
            for t in range(size):
                graph.add_node(
                    f"{i} - t{t}",
                    [output[i][t]],
                    id_=f"{i}-t{t}",
                    metadata={"time": t, "variable": i},
                )

        for t in range(1, size):
            for edge in self.edges:
                if t - edge.lag < 0:
                    logger.debug(
                        f"Cannot generate data for {edge.source} at t={t}, "
                        f"since it depends on {edge.lag}-steps-ago value"
                    )
                else:
                    graph.add_edge(
                        graph.nodes[f"{edge.source}-t{t - edge.lag}"],
                        graph.nodes[f"{edge.target}-t{t}"],
                        metadata={},
                    )
                    graph.remove_directed_edge(
                        graph.nodes[f"{edge.target}-t{t}"],
                        graph.nodes[f"{edge.source}-t{t - edge.lag}"],
                    )

        return output, graph


class IIDSampleGenerator(AbstractSampleGenerator):
    """
    A sample generator that generates data for a sample graph without a time dimension.

    Generators are written as a lambda function that takes two arguments: the current step and the input.
    The input is a SimpleNamespace object that contains the current values of all variables.

    A variable can not depend on itself.


    Example:
    >>> sg = IIDSampleGenerator(
    >>>     initial_values={
    >>>         "Z": random(),
    >>>         "Y": random(),
    >>>         "X": random(),
    >>>     },
    >>>     variables={
    >>>         "param1": 2,
    >>>         "param2": 3,
    >>>     },
    >>>     # generate the dependencies of variables on values of themselves and other variables
    >>>     generators={
    >>>         "Z": lambda s, i: random(),
    >>>         "Y": lambda s, i: i.param1 * i.Z + random(),
    >>>         "X": lambda s, i: i.param2 * i.Y + random()
    >>>     },
    >>>     edges=[
    >>>         SampleEdge("Z", "Y"),
    >>>         SampleEdge("Y", "X"),
    >>>     ]
    >>> )

    """

    # TODO: fix this class. Bug: IIDSampleGenerator only depends on the initial value, it should depend on the step
    def _generate_data(self, size):
        internal_repr = {}

        # Initialize the output dictionary
        for i, v in self.initial_values.items():
            internal_repr[i] = CurrentElementProxy(v)

        # Generate the data for each time step
        for step in range(1, size):
            for value in self.generators.keys():
                generator_input = internal_repr
                generator_input.update(self.vars)
                internal_repr[value].append(
                    self.generators[value](step, SimpleNamespace(**generator_input))
                )
        output = {}
        for i in self.generators.keys():
            output[i] = internal_repr[i].to_list()

        return output

    def generate(self, size):
        """
        Generate data for a sample graph without a time dimension
        :param size: the number of data points to generate
        :return: the generated data and the sample graph
        """
        output = self._generate_data(size)
        graph = Graph()

        for i in self.generators.keys():
            graph.add_node(f"{i}", output[i], id_=f"{i}")

        for edge in self.edges:
            graph.add_edge(
                graph.nodes[f"{edge.source}"], graph.nodes[f"{edge.target}"], value={}
            )
            graph.remove_directed_edge(
                graph.nodes[f"{edge.target}"], graph.nodes[f"{edge.source}"]
            )

        return output, graph



import importlib
from typing import List, Tuple


def unpack_run(args):
    tst = args[0]
    del args[0]
    return tst(*args)


def serialize_module_name(cls):
    return f"{cls.__class__.__module__}.{cls.__class__.__name__}"


def load_pipeline_artefact_by_definition(step):
    name = step["name"]
    path = ".".join(name.split(".")[:-1])
    cls = name.split(".")[-1]
    st_function = importlib.import_module(path)
    st_function = getattr(st_function, cls)
    if not st_function:
        raise ValueError(f"{name} not found")

    if "params" not in step.keys():
        return st_function()
    else:
        return st_function(**step["params"])


def load_pipeline_steps_by_definition(steps):
    pipeline = []
    for step in steps:
        st_function = load_pipeline_artefact_by_definition(step)
        pipeline.append(st_function)
    return pipeline


def retrieve_edges(graph) -> List[Tuple[str, str]]:
    """
    Returns a list of edges from the graph
    :param graph: a graph
    :return: a list of edges
    """
    edges = []
    for u in graph.edges:
        for v in graph.edges[u]:
            edges.append((u, v))
    return edges



import logging
from typing import Tuple, List

from causy.interfaces import (
    PipelineStepInterface,
    TestResult,
    BaseGraphInterface,
    TestResultAction,
)

logger = logging.getLogger(__name__)


class PlaceholderTest(PipelineStepInterface):
    num_of_comparison_elements = 2
    chunk_size_parallel_processing = 10
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> List[TestResult] | TestResult:
        """
        Placeholder test for testing purposes
        :param nodes:
        :param graph:
        :return:
        """
        logger.debug(f"PlaceholderTest {nodes}")
        return TestResult(u=None, v=None, action=TestResultAction.DO_NOTHING, data={})



from causy.interfaces import ExitConditionInterface


class ExitOnNoActions(ExitConditionInterface):
    def check(self, graph, graph_model_instance_, actions_taken, iteration) -> bool:
        """
        Check if there are no actions taken in the last iteration and if so, break the loop
        If it is the first iteration, do not break the loop (we need to execute the first step)
        :param graph: the graph
        :param graph_model_instance_: the graph model instance
        :param actions_taken: the actions taken in the last iteration
        :param iteration: iteration number
        :return: True if you want to break an iteration, False otherwise
        """
        if iteration > 0:
            return True if len(actions_taken) == 0 else False
        else:
            return False



from typing import Tuple

import torch

from causy.generators import AllCombinationsGenerator
from causy.interfaces import (
    PipelineStepInterface,
    ComparisonSettings,
    BaseGraphInterface,
    TestResult,
    TestResultAction,
)


class CalculatePearsonCorrelations(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(self, nodes: Tuple[str], graph: BaseGraphInterface) -> TestResult:
        """
        Calculate the correlation between each pair of nodes and store it to the respective edge.
        :param nodes: list of nodes
        :return: A TestResult with the action to take
        """
        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]
        edge_value = graph.edge_value(graph.nodes[nodes[0]], graph.nodes[nodes[1]])

        x_val = x.values
        y_val = y.values

        cov_xy = torch.mean((x_val - x_val.mean()) * (y_val - y_val.mean()))
        std_x = x_val.std(unbiased=False)
        std_y = y_val.std(unbiased=False)
        pearson_correlation = cov_xy / (std_x * std_y)

        edge_value["correlation"] = pearson_correlation.item()

        return TestResult(
            u=x,
            v=y,
            action=TestResultAction.UPDATE_EDGE,
            data=edge_value,
        )



import enum
from abc import ABC
from dataclasses import dataclass
from typing import List, Optional, Dict, Set, Tuple, Union
from uuid import uuid4
import logging

import torch

from causy.interfaces import (
    BaseGraphInterface,
    NodeInterface,
    TestResultAction,
    TestResult,
    EdgeInterface,
    EdgeTypeInterface,
)

logger = logging.getLogger(__name__)


@dataclass
class Node(NodeInterface):
    """
    A node is a variable in the graph. It has a name, an id and values. The values are stored as a torch.Tensor.
    A node can also have metadata.
    """

    name: str
    id: str
    values: torch.Tensor
    metadata: Dict[str, any] = None

    def __hash__(self):
        return hash(self.id)


class EdgeType(EdgeTypeInterface):
    """
    Edge types are used to define the type of an edge. Currently, there are by default two different edge types:
    - DIRECTED: a directed edge from u to v
    - UNDIRECTED: an undirected edge between u and v
    """

    DIRECTED = "DIRECTED"
    UNDIRECTED = "UNDIRECTED"


@dataclass
class Edge(EdgeInterface):
    """
    An edge is a connection between two nodes. It has a direction defined by the edge_type and metadata.
    A metadata example could be the p-value of a statistical test.
    An edge can be e.g. undirected, directed or bidirected. Which option is available/chosen depends on the algorithm and the current state of the graph.
    By default an edge always has an undirected and a directed edge_type.
    """

    u: NodeInterface
    v: NodeInterface
    edge_type: EdgeType
    metadata: Dict[str, any] = None

    def __hash__(self):
        return hash((self.u.id, self.v.id))


class GraphError(Exception):
    pass


class Graph(BaseGraphInterface):
    """
    The graph represents the internal data structure of causy. It is a simple graph with nodes and edges.
    But it supports to be handled as a directed graph, undirected graph and bidirected graph, which is important to implement different algorithms in different stages.
    It also stores the history of the actions taken on the graph.
    """

    nodes: Dict[str, Node]
    edges: Dict[str, Dict[str, Edge]]
    _reverse_edges: Dict[str, Dict[str, Edge]]
    edge_history: Dict[Tuple[str, str], List[TestResult]]
    action_history: List[Dict[str, List[TestResult]]]

    def __init__(self):
        self.nodes = {}
        self.edges = {}
        self._reverse_edges = {}
        self.edge_history = {}
        self.action_history = []

    def add_edge(self, u: Node, v: Node, metadata: Dict):
        """
        Add an edge to the graph
        :param u: u node
        :param v: v node
        :param metadata: metadata of the edge
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id == v.id:
            raise GraphError("Self loops are currently not allowed")

        if u.id not in self.edges:
            self.edges[u.id] = {}
            self._reverse_edges[u.id] = {}
        if v.id not in self.edges:
            self.edges[v.id] = {}
            self._reverse_edges[v.id] = {}

        a_edge = Edge(u=u, v=v, edge_type=EdgeType.UNDIRECTED, metadata=metadata)
        self.edges[u.id][v.id] = a_edge
        self._reverse_edges[v.id][u.id] = a_edge

        b_edge = Edge(u=u, v=v, edge_type=EdgeType.UNDIRECTED, metadata=metadata)
        self.edges[v.id][u.id] = b_edge
        self._reverse_edges[u.id][v.id] = b_edge

        self.edge_history[(u.id, v.id)] = []
        self.edge_history[(v.id, u.id)] = []

    def add_directed_edge(self, u: Node, v: Node, metadata: Dict):
        """
        Add a directed edge from u to v to the graph
        :param u: u node
        :param v: v node
        :param metadata: metadata of the edge
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id == v.id:
            raise GraphError("Self loops are currently not allowed")

        if u.id not in self.edges:
            self.edges[u.id] = {}
        if v.id not in self._reverse_edges:
            self._reverse_edges[v.id] = {}

        edge = Edge(u=u, v=v, edge_type=EdgeType.DIRECTED, metadata=metadata)

        self.edges[u.id][v.id] = edge
        self._reverse_edges[v.id][u.id] = edge

        self.edge_history[(u.id, v.id)] = []

    def retrieve_edge_history(
        self, u, v, action: TestResultAction = None
    ) -> List[TestResult]:
        """
        Retrieve the edge history
        :param u:
        :param v:
        :param action:
        :return:
        """
        if action is None:
            return self.edge_history[(u.id, v.id)]

        if (u.id, v.id) not in self.edge_history:
            return []

        return [i for i in self.edge_history[(u.id, v.id)] if i.action == action]

    def add_edge_history(self, u, v, action: TestResult):
        """
        Add an action to the edge history
        :param u:
        :param v:
        :param action:
        :return:
        """
        if (u.id, v.id) not in self.edge_history:
            self.edge_history[(u.id, v.id)] = []
        self.edge_history[(u.id, v.id)].append(action)

    def remove_edge(self, u: Node, v: Node):
        """
        Remove an edge from the graph (undirected)
        :param u: u node
        :param v: v node
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id in self.edges and v.id in self.edges[u.id]:
            del self.edges[u.id][v.id]
            del self._reverse_edges[u.id][v.id]

        if v.id in self.edges and u.id in self.edges[v.id]:
            del self.edges[v.id][u.id]
            del self._reverse_edges[v.id][u.id]

    def remove_directed_edge(self, u: Node, v: Node):
        """
        Remove an edge from the graph
        :param u: u node
        :param v: v node
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id not in self.edges:
            return  # no edges from u
        if v.id not in self.edges[u.id]:
            return

        del self.edges[u.id][v.id]
        del self._reverse_edges[v.id][u.id]

    def update_edge(
        self, u: Node, v: Node, metadata: Dict = None, edge_type: EdgeType = None
    ):
        """
        Update an undirected edge in the graph
        :param u: u node
        :param v: v node
        :param value:
        :param edge_type:
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")
        if u.id not in self.edges:
            raise GraphError(f"Node {u} does not have any edges")
        if v.id not in self.edges:
            raise GraphError(f"Node {v} does not have any edges")

        if u.id not in self.edges[v.id]:
            raise GraphError(f"There is no edge from {u} to {v}")

        if v.id not in self.edges[u.id]:
            raise GraphError(f"There is no edge from {v} to {u}")

        if metadata is not None:
            self.edges[u.id][v.id].metadata = metadata
            self.edges[v.id][u.id].metadata = metadata

            self._reverse_edges[u.id][v.id].metadata = metadata
            self._reverse_edges[v.id][u.id].metadata = metadata

        if edge_type is not None:
            self.edges[u.id][v.id].edge_type = edge_type
            self.edges[v.id][u.id].edge_type = edge_type

            self._reverse_edges[u.id][v.id].edge_type = edge_type
            self._reverse_edges[v.id][u.id].edge_type = edge_type

    def update_directed_edge(
        self, u: Node, v: Node, metadata: Dict = None, edge_type: EdgeType = None
    ):
        """
        Update an edge in the graph
        :param u: u node
        :param v: v node
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")
        if u.id not in self.edges:
            raise GraphError(f"Node {u} does not have any edges")
        if v.id not in self.edges[u.id]:
            raise GraphError(f"There is no edge from {u} to {v}")

        if metadata is not None:
            self.edges[u.id][v.id].metadata = metadata
            self._reverse_edges[v.id][u.id].metadata = metadata

        if edge_type is not None:
            self.edges[u.id][v.id].edge_type = edge_type
            self._reverse_edges[v.id][u.id].edge_type = edge_type

    def edge_exists(self, u: Node, v: Node):
        """
        Check if any edge exists between u and v. Cases: u -> v, u <-> v, u <- v
        :param u: node u
        :param v: node v
        :return: True if any edge exists, False otherwise
        """
        if u.id not in self.nodes:
            return False
        if v.id not in self.nodes:
            return False
        if u.id in self.edges and v.id in self.edges[u.id]:
            return True
        if v.id in self.edges and u.id in self.edges[v.id]:
            return True
        return False

    def directed_edge_exists(self, u: Node, v: Node):
        """
        Check if a directed edge exists between u and v. Cases: u -> v, u <-> v
        :param u: node u
        :param v: node v
        :return: True if a directed edge exists, False otherwise
        """
        if u.id not in self.nodes:
            return False
        if v.id not in self.nodes:
            return False
        if u.id not in self.edges:
            return False
        if v.id not in self.edges[u.id]:
            return False

        return True

    def only_directed_edge_exists(self, u: Node, v: Node):
        """
        Check if a directed edge exists between u and v, but no directed edge exists between v and u. Case: u -> v
        :param u: node u
        :param v: node v
        :return: True if only directed edge exists, False otherwise
        """
        if self.directed_edge_exists(u, v) and not self.directed_edge_exists(v, u):
            return True
        return False

    def undirected_edge_exists(self, u: Node, v: Node):
        """
        Check if an undirected edge exists between u and v. Note: currently, an undirected edges is implemented just as
        a directed edge. However, they are two functions as they mean different things in different algorithms.
        Currently, this function is used in the PC algorithm, where an undirected edge is an edge which could not be
        oriented in any direction by orientation rules.
        Later, a cohersive naming scheme should be implemented.
        :param u: node u
        :param v: node v
        :return: True if an undirected edge exists, False otherwise
        """
        if self.directed_edge_exists(u, v) and self.directed_edge_exists(v, u):
            return True
        return False

    def edge_of_type_exists(
        self, u: Node, v: Node, edge_type: EdgeType = EdgeType.DIRECTED
    ):
        """
        Check if an edge of a specific type exists between u and v.
        :param u: node u
        :param v: node v
        :param edge_type: the type of the edge to check for
        :return: True if an edge of this type exists, False otherwise
        """

        if u.id not in self.nodes:
            return False
        if v.id not in self.nodes:
            return False
        if u.id not in self.edges:
            return False
        if v.id not in self.edges[u.id]:
            return False

        if self.edges[u.id][v.id].edge_type != edge_type:
            return False

        return True

    def edge_value(self, u: Node, v: Node) -> Optional[Dict]:
        """
        retrieve the value of an edge
        :param u:
        :param v:
        :return:
        """

        if u.id not in self.edges:
            return None
        if v.id not in self.edges[u.id]:
            return None
        return self.edges[u.id][v.id].metadata

    def edge_type(self, u: Node, v: Node) -> Optional[EdgeType]:
        """
        retrieve the value of an edge
        :param u:
        :param v:
        :return:
        """

        if u.id not in self.edges:
            return None
        if v.id not in self.edges[u.id]:
            return None
        return self.edges[u.id][v.id].edge_type

    def add_node(
        self,
        name: str,
        values: Union[List[float], torch.Tensor],
        id_: str = None,
        metadata: Dict[str, any] = None,
    ) -> Node:
        """
        Add a node to the graph
        :param name: name of the node
        :param values: values of the node
        :param id_: id_ of the node
        :param : node

        :return: created Node
        """
        if id_ is None:
            id_ = str(uuid4())

        if id_ in self.nodes:
            raise ValueError(f"Node with id {id_} already exists")

        if isinstance(values, torch.Tensor):
            tensor_values = values
        else:
            try:
                tensor_values = torch.tensor(values, dtype=torch.float32)
            except TypeError as e:
                raise ValueError(f"Currently only numeric values are supported. {e}")

        if metadata is None:
            metadata = {}

        node = Node(name=name, id=id_, values=tensor_values, metadata=metadata)

        self.nodes[id_] = node
        return node

    def directed_path_exists(self, u: Node, v: Node):
        """
        Check if a directed path from u to v exists
        :param u: node u
        :param v: node v
        :return: True if a directed path exists, False otherwise
        """
        if self.directed_edge_exists(u, v):
            return True
        for w in self.edges[u.id]:
            if self.directed_path_exists(self.nodes[w], v):
                return True
        return False

    def directed_paths(self, u: Node, v: Node):
        """
        Return all directed paths from u to v
        :param u: node u
        :param v: node v
        :return: list of directed paths
        """
        # TODO: try a better data structure for this
        if self.directed_edge_exists(u, v):
            return [[(u, v)]]
        paths = []
        for w in self.edges[u.id]:
            if self.directed_edge_exists(u, self.nodes[w]):
                for path in self.directed_paths(self.nodes[w], v):
                    paths.append([(u, self.nodes[w])] + path)
        return paths

    def parents_of_node(self, u: Node):
        """
        Return all parents of a node u
        :param u: node u
        :return: list of nodes (parents)
        """
        return [self.nodes[n] for n in self._reverse_edges[u.id].keys()]

    def inducing_path_exists(self, u: Node, v: Node):
        """
        Check if an inducing path from u to v exists.
        An inducing path from u to v is a directed path from u to v on which all mediators are colliders.
        :param u: node u
        :param v: node v
        :return: True if an inducing path exists, False otherwise
        """
        if not self.directed_path_exists(u, v):
            return False
        for path in self.directed_paths(u, v):
            for i in range(1, len(path) - 1):
                r, w = path[i]
                if not self.bidirected_edge_exists(r, w):
                    # TODO: check if this is correct (@sof)
                    return True
        return False



import copy
import itertools
import logging
import random

from causy.interfaces import (
    ComparisonSettings,
    GeneratorInterface,
    BaseGraphInterface,
    GraphModelInterface,
    AS_MANY_AS_FIELDS,
)
from causy.graph_utils import load_pipeline_artefact_by_definition

logger = logging.getLogger(__name__)


class AllCombinationsGenerator(GeneratorInterface):
    """
    Generates all combinations of nodes in the graph
    """

    def generate(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        start = self.comparison_settings.min
        # if min is longer then our dataset, we can't create any combinations
        if start > len(graph.nodes):
            return

        # if max is AS_MANY_AS_FIELDS, we set it to the length of the dataset + 1
        if self.comparison_settings.max == AS_MANY_AS_FIELDS:
            stop = len(graph.nodes) + 1
        else:
            stop = self.comparison_settings.max + 1

        # if start is longer then our dataset, we set it to the length of the dataset
        if stop > len(graph.nodes) + 1:
            stop = len(graph.nodes) + 1

        # if stop is smaller then start, we can't create any combinations
        if stop < start:
            return

        # create all combinations
        for r in range(start, stop):
            for i in itertools.combinations(graph.nodes, r):
                yield i


class PairsWithNeighboursGenerator(GeneratorInterface):
    """
    Generates all combinations of pairs of nodes with their neighbours
    """

    shuffle_combinations = True
    chunked = True

    def __init__(
        self,
        comparison_settings: ComparisonSettings,
        chunked: bool = None,
        shuffle_combinations: bool = None,
    ):
        super().__init__(comparison_settings, chunked)
        if shuffle_combinations is not None:
            self.shuffle_combinations = shuffle_combinations

        if chunked is not None:
            self.chunked = chunked

    def generate(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        start = self.comparison_settings.min
        # if min is longer then our dataset, we can't create any combinations
        if start > len(graph.nodes):
            return

        # if max is AS_MANY_AS_FIELDS, we set it to the length of the dataset + 1
        if self.comparison_settings.max == AS_MANY_AS_FIELDS:
            stop = len(graph.nodes) + 1
        else:
            stop = self.comparison_settings.max + 1

        # if start is longer then our dataset, we set it to the length of the dataset
        if stop > len(graph.nodes) + 1:
            stop = len(graph.nodes) + 1

        # if stop is smaller then start, we can't create any combinations
        if stop < start:
            return

        if start < 2:
            raise ValueError("PairsWithNeighboursGenerator: start must be at least 2")

        for i in range(start, stop):
            logger.debug(f"PairsWithNeighboursGenerator: i={i}")
            checked_combinations = set()
            local_edges = copy.deepcopy(graph.edges)
            for node in local_edges:
                for neighbour in local_edges[node]:
                    if (node, neighbour) in checked_combinations:
                        continue

                    checked_combinations.add((node, neighbour))
                    if i == 2:
                        yield (node, neighbour)
                        continue

                    other_neighbours = set(graph.edges[node])
                    if neighbour in other_neighbours:
                        other_neighbours.remove(neighbour)
                    else:
                        continue
                    if len(other_neighbours) + 2 < i:
                        continue
                    combinations = itertools.combinations(other_neighbours, i)
                    if self.shuffle_combinations:
                        combinations = list(combinations)
                        import random

                        random.shuffle(combinations)

                    if self.chunked:
                        chunk = []
                        for k in combinations:
                            chunk.append([node, neighbour] + [ks for ks in k])
                        yield chunk
                    else:
                        for k in combinations:
                            yield [node, neighbour] + [ks for ks in k]


class RandomSampleGenerator(GeneratorInterface):
    """
    Executes another generator and returns a random sample of the results
    """

    every_nth = 100

    def __init__(
        self,
        comparison_settings: ComparisonSettings = None,
        chunked: bool = None,
        every_nth: int = None,
        generator: GeneratorInterface = None,
    ):
        super().__init__(comparison_settings, chunked)
        if every_nth is not None:
            self.every_nth = every_nth

        if generator is not None:
            if isinstance(generator, GeneratorInterface):
                self.generator = generator
            else:
                self.generator = load_pipeline_artefact_by_definition(generator)
        else:
            raise ValueError("RandomSampleGenerator: generator must be set")

    def generate(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        """
        Executes another generator and returns a random sample of the results
        :param graph:
        :param graph_model_instance_:
        :return: yields a random sample of the results
        """
        for combination in self.generator.generate(graph, graph_model_instance_):
            if random.randint(0, self.every_nth) == 0:
                yield combination



"""causy

Causal discovery made easy. Causy allows you to use and implement causal discovery algorithms with easy to use, extend and maintain pipelines. It is built based on pytorch which allows you to run the algorithms on CPUs as well as GPUs seamlessly.

Learn more at https://causy.dev.

"""

import importlib.metadata

__version__ = importlib.metadata.version("causy")



from typing import Tuple, List, Optional

from causy.generators import AllCombinationsGenerator
from causy.interfaces import (
    TestResultAction,
    PipelineStepInterface,
    ComparisonSettings,
    BaseGraphInterface,
    TestResult,
)


class ColliderRuleFCI(PipelineStepInterface):
    generator = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing = 1
    parallel = False

    def test(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> Optional[List[TestResult] | TestResult]:
        """
        Some notes on how we implment FCI: After the independence tests, we have a graph with undirected edges which are
        implemented as two directed edges, one in each direction. We initialize the graph by adding values to all these edges,
        in the beginning, they get the value "either directed or undirected". Then we perform the collider test. Unlike in PC,
        we do not delete directed edges from z to u and from z to v in order to obtain the structure (u -> z <- v). Instead, we
        delete the information "either directed or undirected" from the directed edges from u to z and from v to z. That means,
        the directed edges from u to z and from v to z are now truly directed edges. The edges from z to u and from z to v can
        still stand for a directed edge or no directed edge. In the literature, this is portrayed by the meta symbol * and we
        obtain u *-> z <-* v. There might be ways to implement these similar but still subtly different orientation rules more consistently.

        TODO: write tests

        We call triples u, v, z of nodes v structures if u and v that are NOT adjacent but share an adjacent node z.
        V structures looks like this in the undirected skeleton: (u - z - v).
        We now check if z is in the separating set. If so, the edges must be oriented from u to z and from v to z:
        (u *-> z <-* v), where * indicated that there can be an arrowhead or none, we do not know, at least until
        applying further rules.
        :param nodes: list of nodes
        :param graph: the current graph
        :returns: list of actions that will be executed on graph
        """

        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]

        # if u and v are adjacent, do nothing
        if graph.undirected_edge_exists(x, y):
            return TestResult(u=x, v=y, action=TestResultAction.DO_NOTHING, data={})

        # if u and v are NOT adjacent, store all shared adjacent nodes
        potential_zs = set(graph.edges[x.id].keys()).intersection(
            set(graph.edges[y.id].keys())
        )

        actions = graph.retrieve_edge_history(
            x, y, TestResultAction.REMOVE_EDGE_UNDIRECTED
        )

        # if u and v are not independent given z, safe action: make z a collider
        results = []
        for z in potential_zs:
            z = graph.nodes[z]

            separators = []
            for action in actions:
                if "separatedBy" in action.data:
                    separators += [a.id for a in action.data["separatedBy"]]

            if z.id not in separators:
                results += [
                    TestResult(
                        u=x,
                        v=z,
                        action=TestResultAction.UPDATE_EDGE_DIRECTED,
                        data={"edge_type": None},
                    ),
                    TestResult(
                        u=y,
                        v=z,
                        action=TestResultAction.UPDATE_EDGE_DIRECTED,
                        data={"edge_type": None},
                    ),
                ]
        return results



import os
from datetime import datetime

import fastapi
import uvicorn
from typing import Any, Dict, List, Optional

from fastapi import APIRouter
from pydantic import BaseModel, Json, UUID4, Field
from starlette.staticfiles import StaticFiles

import logging

logger = logging.getLogger(__name__)

API_ROUTES = APIRouter()

MODEL = None


class CausyAlgorithm(BaseModel):
    type: str
    reference: str


class CausyNodePosition(BaseModel):
    x: Optional[float]
    y: Optional[float]


class CausyNode(BaseModel):
    id: UUID4
    name: str
    position: Optional[CausyNodePosition] = None


class CausyEdgeValue(BaseModel):
    metadata: Dict[str, Any] = None
    edge_type: str = None


class CausyEdge(BaseModel):
    class Config:
        populate_by_name = True

    from_field: CausyNode = Field(alias="from")
    to: CausyNode
    value: CausyEdgeValue


class CausyModel(BaseModel):
    name: str
    created_at: datetime
    algorithm: CausyAlgorithm
    steps: List[Dict[str, Any]]
    nodes: Dict[UUID4, CausyNode]
    edges: List[CausyEdge]


@API_ROUTES.get("/status", response_model=Dict[str, Any])
async def get_status():
    """Get the current status of the API."""
    return {"status": "ok"}


@API_ROUTES.get("/model", response_model=CausyModel)
async def get_model():
    """Get the current model."""
    return MODEL


def server(result: Dict[str, Any]):
    """Create the FastAPI server."""
    app = fastapi.FastAPI(
        title="causy-api",
        version="0.0.1",
        description="causys internal api to serve data from the result files",
    )
    global MODEL
    MODEL = CausyModel(**result)

    app.include_router(API_ROUTES, prefix="/api/v1", tags=["api"])
    app.mount(
        "",
        StaticFiles(
            directory=os.path.join(os.path.dirname(__file__), "static"), html=True
        ),
        name="static",
    )

    host = os.getenv("HOST", "localhost")
    port = int(os.getenv("PORT", "8000"))
    cors_enabled = os.getenv("CORS_ENABLED", "false").lower() == "true"

    # cors e.g. for development of separate frontend
    if cors_enabled:
        logger.warning("🌐 CORS enabled")
        from fastapi.middleware.cors import CORSMiddleware

        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    server_config = uvicorn.Config(app, host=host, port=port, log_level="error")
    server = uvicorn.Server(server_config)
    return server_config, server



import logging
from scipy import stats as scipy_stats
import math


logger = logging.getLogger(__name__)


def sum_lists(*lists):
    """
    :param lists: lists of numbers
    :return: list (sum of lists)
    """
    return list(map(sum, zip(*lists)))


def get_t_and_critical_t(sample_size, nb_of_control_vars, par_corr, threshold):
    # TODO: rewrite this function with torch
    # check if we have to normalize data
    deg_of_freedom = sample_size - 2 - nb_of_control_vars
    if abs(round(par_corr, 4)) == 1:
        return (1, 0)
    critical_t = scipy_stats.t.ppf(1 - threshold / 2, deg_of_freedom)
    t = par_corr * math.sqrt(deg_of_freedom / (1 - par_corr**2))
    return t, critical_t
